#ifndef COLUMN_H_INCLUDED
#define COLUMN_H_INCLUDED

#include "AllUsefulHeaders.hpp"

/// Class 'column' for a matrix
// class 'list of pointers to reals' == "smart" column
class Column : public list<Real*> {

public :


	/*****************************
	* constructors, destructors
	******************************/

	Column () {}

	// Constructor with NULL pointers initialization
	Column (Int nbElems) {
		for (Int i=nbElems-1; i>=0; i--)
			push_back(NULL);
	}

	Column (const Column& col) {
		for (list<Real*>::const_iterator it=col.begin(); it!=col.end(); it++)
			push_back(*it);
	}

	// only if the columns came from a substraction :
	void freeElems() {
		for (list<Real*>::iterator it=begin(); it!=end(); it++)
			free(*it);
	}

	// nothing to do : values freed in rows
	~Column () {
		clear();
	}


	/*****************************
	* operators definitions
	******************************/

	inline void operator=(Column& col) {
		list<Real*>::iterator it1=begin();
		list<Real*>::iterator it2=col.begin();
		while (it1!=end()) {
			*(*it1) = *(*it2);
			it1++;
			it2++;
		}
	}

	inline void operator+=(Column& col) {
		list<Real*>::iterator it1=begin();
		list<Real*>::iterator it2=col.begin();
		while (it1!=end()) {
			*(*it1) += *(*it2);
			it1++;
			it2++;
		}
	}

	inline void operator-=(Column& col) {
		list<Real*>::iterator it1=begin();
		list<Real*>::iterator it2=col.begin();
		while (it1!=end()) {
			*(*it1) -= *(*it2);
			it1++;
			it2++;
		}
	}
	
	inline Column operator-(Column col) {
		Column result(size());
		Real** vals = (Real**)calloc(size(),sizeof(Real*));
		Int countV=0;
		list<Real*>::iterator it1=begin(),
							it2 = col.begin();
		for (list<Real*>::iterator it=result.begin(); it!=result.end(); it++) {
			vals[countV] = (Real*)malloc(sizeof(Real));
			*it = vals[countV++];
			*(*it) = *(*it1) - *(*it2);
			it1++;
			it2++;
		}
		free(vals);
		return result;
	}

	inline void operator*=(Real scal) {
		for (list<Real*>::iterator it=begin(); it!=end(); it++)
			*(*it) *= scal;
	}

	inline void operator/=(Real scal) {
		for (list<Real*>::iterator it=begin(); it!=end(); it++)
			*(*it) /= scal;
	}

	/// Function to zero all column entries
	inline void zero() {
		for (list<Real*>::iterator it=begin(); it!=end(); it++)
			*(*it) = 0.0;
	}
	
	/// This function replaces each element c by c^n in the column
	inline void operator^=(U_Int n) {
		for (list<Real*>::iterator it=begin(); it!=end(); it++)
			*(*it) = puiss(*(*it),n);
	}

	
	/**********************
	* Fonctions diverses
	**********************/

	/// Returns the sum all the elements
	inline Real sum() {
		Real result=0.0;
		for (list<Real*>::iterator it=begin(); it!=end(); it++)
			result += *(*it);
		return result;
	}

	/// Returns the average column value
	inline Real mean() {
		if (empty()) return 0.0;
		return sum() / size();
	}

	/// Computes euclidian squared norm
	inline Real sq_norm2() {
		if (empty()) return 0.0;
		Real result=0.0;
		for (list<Real*>::iterator it=begin(); it!=end(); it++)
			result += *(*it) * *(*it);
		return result;
	}

	/// Computes euclidian norm
	inline Real norm2() {
		return sqrt(sq_norm2());
	}

	/// Normalizes the column to 1 for the euclidian norm
	inline void normalize2() {
		(*this) /= norm2();
	}

	/// Computes the euclidian squared distance to another column
	inline Real sq_distance2(Column& col) {
		Column tmp = *this;
		tmp -= col;
		return tmp.sq_norm2();
	}

	/// Compute the euclidian distance to another column
	inline Real distance2(Column& col) {
		return sqrt(this->sq_distance2(col));
	}

	/// Computes the scalar product with another column
	inline Real operator*(Column& col) {
		if (empty()) return 0.0;
		Real result=0.0;
		list<Real*>::iterator it1=begin();
		list<Real*>::iterator it2=col.begin();
		while (it1!=end()) {
			result += *(*it1) * *(*it2);
			it1++;
			it2++;
		}
		return result;
	}

	/// remove projection onto some other column (orthogonalisation process)
	// a := a - [ <a,b> / ||b||^2 ] b
	inline void removeProjectionOnto(Column& col) {
		Real coeff= (*this) * col, // <a,b>
			factNorm=col.sq_norm2(); // / ||b||^2
		if (factNorm>0.0) coeff/=factNorm;
		Column tmp = col;
		tmp *= coeff; // coeff * b
		(*this) -= tmp; // a - coeff * b
	}
};

#endif // Column
