#ifndef MATRICE_H_INCLUDED
#define MATRICE_H_INCLUDED

#include "AllUsefulHeaders.hpp"
#include "Vecteur.hpp"
#include "Column.hpp"

class Matrice {

private :

	MyVector<Vecteur> rows; ///< Rows of the object
	vector<Column> cols; ///< corresponding columns

	// assumption : rows already built (at least row 0)
	inline void initializeCols() {
		for (Int i=rows[0].nbElems-1; i>=0; i--) {
			Column tmpCol;
			cols.push_back(tmpCol);
		}
	}

	// build columns by taking all rows elements addresses ;
	// not called very often thanks to drop shot complexity
	inline void buildColsFromRows() {
		for (Int i=nbCols()-1; i>=0; i--)
			cols[i].clear();
		for (Int i=nbCols()-1; i>=0; i--) {
			for (vector<Vecteur>::iterator it=rows.begin(); it!=rows.end(); it++)
				cols[i].push_back(&((*it)[i]));
		}
	}

	// just add columns values (rows not reallocated)
	inline void adjustColsFromRows(Int deltaNbRows) { // deltaNbRows = nbRows - oldNbRows
		// here cols is never empty, the size difference will be nbRows - oldNbRows
		if (deltaNbRows <= 0) {
			// easy case : just pop_back
			for (Int i=nbCols()-1; i>=0; i--) {
				for (Int j=deltaNbRows+1; j<=0; j++)
					cols[i].pop_back();
			}
		}
		else {
			// last case : push_back some values
			for (Int i=nbCols()-1; i>=0; i--) {
				for (Int j=0; j<deltaNbRows; j++)
					cols[i].push_back(&(rows[nbRows()-1+j][i]));
			}
		}
	}

public :

	inline U_Int nbRows() {
		return rows.size();
	}

	inline U_Int nbCols() {
		return cols.size();
	}

	inline vector<Vecteur>::iterator beginRows() {
		return rows.begin();
	}

	inline vector<Vecteur>::iterator endRows() {
		return rows.end();
	}

	inline vector<Column>::iterator beginCols() {
		return cols.begin();
	}

	inline vector<Column>::iterator endCols() {
		return cols.end();
	}

	inline bool empty() {
		return rows.empty();
	}

	inline void clear() {
		rows.clear();
		cols.clear();
	}

	/// returns the i-th row (Vecteur, or Fonction)
	inline Vecteur& operator[](Int i) {
		return rows[i];
	}

	/// returns the i-th column
	inline Column& getCol(Int i) {
		return cols[i];
	}


	/********************************
	* constructeurs, destructeurs :
	********************************/

	Matrice () {}

	/// initialize rows and columns to given sizes, with random values
	Matrice (Int nbRows, Int nbCols, Int lastLeftIndex = -1) {
		for (Int i=0; i<nbRows; i++) {
			Vecteur tmpRow(nbCols,lastLeftIndex);
			rows.push_back(tmpRow);
		}
		initializeCols();
		buildColsFromRows();
	}

	inline void setToZero() {
		for (Int i=rows.size()-1; i>=0; i--)
			rows[i].zero();
	}

	/// this = I (if not square, then intuitive result)
	inline void setToIdentity() {
		setToZero();
		for (Int i=rows.size()-1; i>=0; i--) {
			if (i < (Int)nbCols()) rows[i][i] = 1.0;
		}
	}

	/// add a new row
	inline void addRow(Vecteur& v_f) {
		bool reallocRows = rows.addElem(v_f);
		if (cols.empty()) initializeCols();
		if (reallocRows) buildColsFromRows();
		else adjustColsFromRows(1);
	}

	/// Build rows from m_f's rows which are marked in 'indices'
	Matrice (Matrice& m_f, vector<Int>& indices) {
		for (vector<Int>::iterator it=indices.begin(); it!=indices.end(); it++)
			addRow(m_f[*it]);
	}

	/// Removes the last row
	inline void removeLastRow() {
		rows.pop_back();
		for (vector<Column>::iterator it=cols.begin(); it!=cols.end(); it++)
			(*it).pop_back();
	}

	/// delete the row pointed by some iterator
	inline void eraseRow(vector<Vecteur>::iterator& it) {
		// get the index corresponding to it:
		Int indxIt = 0;
		vector<Vecteur>::iterator itI=rows.begin();
		while (itI!=rows.end()) {
			if (itI == it) break;
			indxIt++;
			itI++;
		}

		// erase row :
		rows.erase(it);

		// erase corresponding column values :
		for (Int i=nbCols()-1; i>=0; i--) {
			list<Real*>::iterator itC=cols[i].begin();
			for (Int j=indxIt-1; j>=0; j--) itC++;
			cols[i].erase(itC);
		}
	}

	/// copy constructor, sometimes truncated to the k first cols
	Matrice (const Matrice& m_f, Int nbCols=0) {
		for (vector<Vecteur>::const_iterator it=m_f.rows.begin(); it!=m_f.rows.end(); it++) {
			Vecteur tmpRow(*it,nbCols);
			addRow(tmpRow);
		}
	}

	/// constructor which concatenates m_f1 and m_f2 in rows
	Matrice (Matrice& m_f1, Matrice& m_f2) {
		for (vector<Vecteur>::iterator it=m_f1.rows.begin(); it!=m_f1.rows.end(); it++)
			addRow(*it);
		for (vector<Vecteur>::iterator it=m_f2.rows.begin(); it!=m_f2.rows.end(); it++)
			addRow(*it);
	}

	~Matrice () {
		rows.clear();
		cols.clear();
	}


	/********************************
	* definitions d'operateurs :
	********************************/

	inline void operator=(Matrice m_f) {
		if (nbCols() != m_f.nbCols()) {
			// worst scenario : we must add columns ; this is not done often
			// useful especially if 'this' is empty
			rows.clear();
			cols.clear();
			for (vector<Vecteur>::iterator it=m_f.beginRows(); it!=m_f.endRows(); it++) rows.push_back(*it);
			initializeCols();
			buildColsFromRows();
		}
		else {
			// "normal scenario" when 'this' isn't empty
			Int deltaNbRows = m_f.nbRows() - nbRows();
			Int lastThis=0;
			if (deltaNbRows <= 0) {
				for (Int i=deltaNbRows+1; i<=0; i++) rows.pop_back();
				lastThis = nbRows();
			}
			else {
				lastThis = nbRows();
				for (Int i=0; i<deltaNbRows; i++) addRow(m_f[i+lastThis]);
			}
			for (Int i=lastThis-1; i>=0; i--) rows[i] = m_f[i];
		}
	}

	inline Matrice operator+(Matrice m_f) {
		Matrice result;
		vector<Vecteur>::iterator it1=rows.begin(),
							it2=m_f.rows.begin();
		while (it1!=rows.end()) {
			Vecteur tmpRow = *it1 + *it2;
			result.addRow(tmpRow);
			it1++;
			it2++;
		}
		return result;
	}

	inline Matrice operator-(Matrice m_f) {
		Matrice result;
		vector<Vecteur>::iterator it1=rows.begin(),
							it2=m_f.rows.begin();
		while (it1!=rows.end()) {
			Vecteur tmpRow = *it1 - *it2;
			result.addRow(tmpRow);
			it1++;
			it2++;
		}
		return result;
	}

	inline void operator+=(Matrice& m_f) {
		vector<Vecteur>::iterator it1=rows.begin();
		vector<Vecteur>::iterator it2=m_f.rows.begin();
		while (it1!=rows.end()) {
			*it1 += *it2;
			it1++;
			it2++;
		}
	}

	inline void operator-=(Matrice& m_f) {
		vector<Vecteur>::iterator it1=rows.begin();
		vector<Vecteur>::iterator it2=m_f.rows.begin();
		while (it1!=rows.end()) {
			*it1 -= *it2;
			it1++;
			it2++;
		}
	}

	inline void operator*=(Real scal) {
		for (vector<Vecteur>::iterator it=rows.begin(); it!=rows.end(); it++)
			*it *= scal;
	}

	inline void operator/=(Real scal) {
		for (vector<Vecteur>::iterator it=rows.begin(); it!=rows.end(); it++)
			*it /= scal;
	}


	/**************************************
	* diverses operations matricielles :
	***************************************/

	/// returns transpose(this) :
	inline Matrice transpose() {
		Matrice thisTr(nbCols(),nbRows());
		for (Int i=0; i<(Int)nbCols(); i++) {
			Int countElems=0;
			for (list<Real*>::iterator it=cols[i].begin(); it!=cols[i].end(); it++)
				thisTr[i][countElems++] = *(*it);
		}
		return thisTr;
	}

	/// matrix * {matrix or functions} product
	template <class MatxOrFuncs>
	inline MatxOrFuncs operator*(MatxOrFuncs& m_f) {
		MatxOrFuncs result(nbRows(),m_f.nbCols());
		Int countRows=0;
		for (vector<Vecteur>::iterator it1=rows.begin(); it1!=rows.end(); it1++) {
			Int countCols=0;
			for (vector<Column>::iterator it2=m_f.cols.begin(); it2!=m_f.cols.end(); it2++)
				result[countRows][countCols++] = *it1 * *it2;
			countRows++;
		}
		return result;
	}

	/// matrix-vector product :
	inline Vecteur operator*(Vecteur& vect) {
		Vecteur result(vect.nbElems);
		for (Int i=0; i<vect.nbElems; i++)
			result[i] = rows[i] * vect;
		return result;
	}

	/// transpose(matrix) * other_matrix product
	inline Matrice trMultBy(Matrice& matx) {
		Matrice result(nbCols(),matx.nbCols());
		Int countRows=0;
		for (vector<Column>::iterator it1=cols.begin(); it1!=cols.end(); it1++) {
			Int countCols=0;
			for (vector<Column>::iterator it2=matx.cols.begin(); it2!=matx.cols.end(); it2++)
				result[countRows][countCols++] = *it1 * *it2;
			countRows++;
		}
		return result;
	}

	/// returns transpose(this) * vect
	inline Vecteur trMultByVect(Vecteur& vect) {
		Vecteur result(nbCols());
		for (Int i=nbCols()-1; i>=0; i--)
			result[i] = vect * cols[i];
		return result;
	}

	/// computes this = this * H for H Householder matrix
	inline void rightMultByH(Vecteur& hVect) {
		// compute this = this * H for H Householder matrix from hVect :
		// "this -= 2 * (this * hVect) * t(hVect)" for block on right,
		// no effect for blocks on left

		/*
		this = ( B1 | B2 )  ,  H= ( I |      0      )
			   ( B3 | B4 )		  ( 0 |  I-2 v t(v) )

		==> compute this * H = ( B1 | B2 - 2B2 v t(v) )
							   ( B3 | B4 - 2B4 v t(v) )
		*/

		Int deltaRow = nbRows() - hVect.nbElems,
		deltaCol = nbCols() - hVect.nbElems;
		Real* scalProdsUp=(Real*)calloc(MAX(0,deltaRow),sizeof(Real)); // B2*v
		Real* scalProdsDown=(Real*)calloc(hVect.nbElems,sizeof(Real)); // B4*v
		for (Int i=nbRows()-1; i>=0; i--) {
			for (Int j=deltaCol; j<(Int)nbCols(); j++) {
				if (i<deltaRow) scalProdsUp[i] += rows[i][j] * hVect[j-deltaCol];
				else scalProdsDown[i-deltaRow] += rows[i][j] * hVect[j-deltaCol];
			}
		}

		// remove those weighted scalar products from B2 and B4 :
		for (Int i=nbRows()-1; i>=0; i--) {
			for (Int j=deltaCol; j<(Int)nbCols(); j++) {
				if (i<deltaRow) rows[i][j] -= 2 * scalProdsUp[i] * hVect[j-deltaCol];
				else rows[i][j] -= 2 * scalProdsDown[i-deltaRow] * hVect[j-deltaCol];
			}
		}
		free(scalProdsUp);
		free(scalProdsDown);
	}

	/// computes this = H * this for H Householder matrix
	inline void leftMultByH(Vecteur& hVect) {
		// compute this = H *this for H Householder matrix from hVect :
		// "this -= 2 * hVect * (t(hVect) * this)" for block on right,
		// no effect for blocks on top

		/*
		this = ( B1 | B2 )  ,  H= ( I |      0      )
			   ( B3 | B4 )		  ( 0 |  I-2 v t(v) )

		==> compute H * this = (       B1        |       B2        )
							   ( B3 - 2v t(v) B3 | B4 - 2v t(v) B4 )
		*/

		Int deltaRow = nbRows() - hVect.nbElems,
		deltaCol = nbCols() - hVect.nbElems;
		Real* scalProdsLeft=(Real*)calloc(MAX(0,deltaCol),sizeof(Real)); // t(v)*B3
		Real* scalProdsRight=(Real*)calloc(hVect.nbElems,sizeof(Real)); // t(v)*B4
		for (Int i=nbCols()-1; i>=0; i--) {
			for (Int j=deltaRow; j<(Int)nbRows(); j++) {
				if (i<deltaCol) scalProdsLeft[i] += rows[j][i] * hVect[j-deltaRow];
				else scalProdsRight[i-deltaCol] += rows[j][i] * hVect[j-deltaRow];
			}
		}

		// remove those weighted scalar products from B3 and B4 :
		for (Int i=nbCols()-1; i>=0; i--) {
			for (Int j=deltaRow; j<(Int)nbRows(); j++) {
				if (i<deltaCol) rows[j][i] -= 2 * scalProdsLeft[i] * hVect[j-deltaRow];
				else rows[j][i] -= 2 * scalProdsRight[i-deltaCol] * hVect[j-deltaRow];
			}
		}
		free(scalProdsLeft);
		free(scalProdsRight);
	}

	/// matrix/functions * transpose(other_matrix/functions) product :
	inline Matrice multByTr(Matrice& m_f) {
		Matrice result(nbRows(),m_f.nbRows());
		for (Int i=rows.size()-1; i>=0; i--) {
			for (Int j=m_f.rows.size()-1; j>=0; j--)
				result[i][j] = rows[i] * m_f.rows[j];
		}
		return result;
	}

	/// compute the "scalar product matrix" = X * tr(X)
	inline Matrice X_trX() {
		Matrice result;
		Vecteur tmpRow(nbRows());
		for (vector<Vecteur>::iterator it1=rows.begin(); it1!=rows.end(); it1++) {
			Int countElems=0;
			for (vector<Vecteur>::iterator it2=rows.begin(); it2!=rows.end(); it2++)
				tmpRow[countElems++] = *it1 * *it2;
			result.addRow(tmpRow);
		}
		return result;
	}


	/***********************
	* fonctions diverses :
	***********************/

	/// normalize all rows to 1
	inline void normalizeRows() {
		for (vector<Vecteur>::iterator it=rows.begin(); it!=rows.end(); it++)
			(*it).normalize2();
	}

	/// computes the mean value of a row
	inline void computeMeanRow(Vecteur& v_f) {
		v_f.zero();
		if (empty()) return;
		for (vector<Vecteur>::iterator it=rows.begin(); it!=rows.end(); it++)
			v_f += *it;
		v_f /= nbRows();
	}

	/// computes the mean row, restricted to some indices
	inline void computePartialMeanRow(vector<Int>& indices,Vecteur& v_f) {
		v_f.zero();
		if (indices.empty()) return;
		for (Int i=indices.size()-1; i>=0; i--)
			v_f += rows[indices[i]];
		v_f /= indices.size();
	}

	// normalisation of distances between rows (> 1 row..) :
	Real meanDistance() {
		Real meanDist = 0.0;
		for (Int i=nbRows()-1; i>=0; i--) {
			for (Int j=nbRows()-1; j>i; j--)
				meanDist += rows[i].distance2(rows[j]);
		}
		return (2*meanDist) / (nbRows()*(nbRows()-1));
	}

	// find average squared distance :
	Real meanSqDistance() {
		Real meanSqDist = 0.0;
		for (Int i=nbRows()-1; i>=0; i--) {
			for (Int j=nbRows()-1; j>i; j--)
				meanSqDist += rows[i].sq_distance2(rows[j]);
		}
		return (2*meanSqDist) / (nbRows()*(nbRows()-1));
	}

	/// scale data : center and standardize
	Matrice scale(Matrice* refScales=NULL) {
		// scaling data : center and standardize;
		Matrice scales(2,nbCols());

		if (refScales==NULL) {
			scales.setToZero();
			for (Int i=nbCols()-1; i>=0; i--) {
				for (Int j=nbRows()-1; j>=0; j--) {
					scales[0][i]+=rows[j][i];
					scales[1][i]+=rows[j][i]*rows[j][i];
				}
				scales[0][i]/=nbRows();
				scales[1][i]=sqrt(scales[1][i]/nbRows()-scales[0][i]*scales[0][i]);
			}
		}
		else scales = *refScales;

		// effective scale of data: (assumption : variance isn't zero, after some variables selection
		for (Int i=nbCols()-1; i>=0; i--) {
			for (Int j=nbRows()-1; j>=0; j--)
				rows[j][i]=(rows[j][i]-scales[0][i])/scales[1][i];
		}
		return scales;
	}


	/**************************
	* read / write to file :
	**************************/

	/// write Matrice to file
	void operator>>(ofstream& outFile) {
		// <value> <separator> <value> <separator> ... <value> \n
		for (Int i=0; i<(Int)nbRows(); i++) {
			for (Int j=0; j<(Int)nbCols()-1; j++)
				outFile << rows[i][j] << " ";
			outFile << rows[i][nbCols()-1] << endl;
		}
	}

	/// read Matrice from file (for any separator)
	void operator<<(ifstream& inFile) {
		string line;
		while (getline(inFile,line,'\n')) {
			// parse the line to get only numerics :
			vector<Real> tmpValues;
			string::iterator it1=line.begin(), it2;
			while (!isEndLine(it1,line)) {
				// put it1 on the next readable digit
				while (!isEndLine(it1,line)
						&& !(*it1=='.' || *it1=='e' || *it1=='E' || *it1=='-'
						|| (*it1 >= '0' && *it1 <= '9')) )
					it1++;
				// put it2 on next separator :
				it2=it1;
				while (!isEndLine(it2,line)
						&& (*it2=='.' || *it2=='e' || *it2=='E' || *it2=='-'
						|| (*it2 >= '0' && *it2 <= '9')) )
					it2++;
				// get the corresponding value :
				string newValue(it1,it2);
				if (!newValue.empty()) tmpValues.push_back(atof(newValue.c_str()));
				// reinitialize it1 to the first next separator character :
				it1=it2;
			}
			if (!tmpValues.empty()) {
				Vecteur tmpRow(tmpValues.size());
				for (Int i=tmpValues.size()-1; i>=0; i--)
					tmpRow[i]=tmpValues[i];
				addRow(tmpRow);
			}
		}
	}

};

#endif // Matrice
