#include "SingularDecomp.hpp"

// solves X beta ~= Y with SVD for least squares ; return beta
Matrice solveLinearReg(Matrice& X,Matrice& Y) {

	// best numerical way : SVD, X = U diag(D) t(V)
	Matrice U(X.nbRows(),X.nbRows());
	Matrice V(X.nbCols(),X.nbCols());
	Int sd = MIN(X.nbRows(),X.nbCols());
	Vecteur D(sd);
	SVD_decomp(X,U,V,D);

    // il faut restreindre U a ses sd=D.nbElems premieres colonnes :
	Matrice newU;
	for (Int i=0; i<(Int)U.nbRows(); i++) {
		Vecteur newRow(sd);
		for (Int j=sd-1; j>=0; j--) newRow[j]=U[i][j];
		newU.addRow(newRow);
	}
	U = newU;

	// puis "diviser" V par D
	for (Int i=sd-1; i>=0; i--) {
        if (D[i] >= ZERO_EPS) V.getCol(i) /= D[i];
        else V.getCol(i).zero();
    }
    Matrice tmpProd = U.trMultBy(Y);
    Matrice beta = V * tmpProd;
	return beta;
}

// decomposition SVD :
void SVD_decomp(Matrice& A, Matrice& U, Matrice& V, Vecteur& D) {
	// Given the matrix A, this routine computes its singular value decomposition,
	// A = U D tV and stores the results in the matrices U and V, and the vector D.

	// PREPROCESS if more columns than rows : SVD of tr(A), and finally just reverse U and V.
	bool transposeA=false;
	if (A.nbRows() < A.nbCols()) {
		A = A.transpose();
		SWAP(U,V); // to get right sizes
		transposeA=true;
	}

	//////////////////////
	// PHASE 1: BIDIAG ///
	/////////////////////

	// first step: Householder bidiagonalization of A <--> U D-W tV
	U.setToIdentity();
	V.setToIdentity();
	Int nbRows = A.nbRows(), nbCols = A.nbCols();

    // shortcut :
	if (nbRows<=1 && nbCols<=1) {
	    D[0]=A[0][0];
	    return;
	}

	Int indLine=0, indCol=0;
	while (indLine<nbRows-1 && indCol<nbCols-1) {
		// find "column" householder vector of size m-indLine
		Vecteur vCol(nbRows-indLine);
		findLeftHouseholdVect(A,vCol,indLine,indCol);

		// we apply left transformation to matrix U AND to matrix A: A:=H*A
		applyLeftTransformation(A,U,vCol,indLine);

		// counters' incrementation
		indCol++;

		// find "line" householder vector of size n-indCol, ONLY IF indCol<n-1
		if (indCol<nbCols-1) {
			Vecteur vLine(nbCols-indCol);
			findRightHouseholdVect(A,vLine,indLine,indCol);

			// we apply right transformation to matrix V AND to matrix A
			applyRightTransformation(A,V,vLine,indCol);
		}

		// counters' incrementation
		indLine++;
	}

	if (indLine<nbRows-1) {
		// means we stopped because not enough columns;
		// the last column remains to be treated
		// find "column" householder vector of size m-indLine
		Vecteur vCol(nbRows-indLine);
		findLeftHouseholdVect(A,vCol,indLine,indCol);

		// we apply left transformation to matrix U AND to matrix A
		applyLeftTransformation(A,U,vCol,indLine);
	}

	if (indCol<nbCols-1) {
		// means we stopped because not enough lines;
		// the last line remains to be treated
		Vecteur vLine(nbCols-indCol);
		findRightHouseholdVect(A,vLine,indLine,indCol);

		// we apply right transformation to matrix V AND to matrix A
		applyRightTransformation(A,V,vLine,indCol);
	}


	//////////////////////
	// PHASE 2: DIAG /////
	/////////////////////

	// second step: B convergence to diagonal form using Givens transformations,
	// remembering the multiplication matrix too (in U and V).
	// rotation matrix : (  cs sn )
	//					 ( -sn cs )
	Int sd=D.nbElems;
	Vecteur surDiag(sd-1);
	for (Int i=sd-2; i>=0; i--) surDiag[i]=A[i][i+1];
	for (Int i=sd-1; i>=0; i--) D[i]=A[i][i];

	while (surDiag.normInf() >= 5.0e-4) {

		Real oldcs = 1.0, oldsn = 0.0;
		Real cs = 1.0, sn=0.0, r=0.0, h=0.0, f, g;
		f = D[0];
		g = surDiag[0];

		// main loop :
		for (Int i = 0; i<sd-1; i++) {
			rot ( f , g, cs, sn, r);

			{
				// do operations on V : V = V * tr(rotMatrix)
				for (Int j=nbCols-1; j>=0; j--) {
					Real tmp = V[j][i];
					V[j][i] = tmp * cs + V[j][i+1] * sn;
					V[j][i+1] = -tmp * sn + V[j][i+1] * cs;
				}
			}

			if (i>=1) surDiag[i-1] = oldsn*r;
			f = oldcs*r;
			g = D[i+1]*sn;
			h = D[i+1]*cs;
			rot(f,g,cs,sn,r);

			{
				// do operations on U : U = U * tr(rotMatrix)
				for (Int j=nbRows-1; j>=0; j--) {
					Real tmp = U[j][i];
					U[j][i] = tmp * cs + U[j][i+1] * sn;
					U[j][i+1] = -tmp * sn + U[j][i+1] * cs;
				}
			}

			D[i]=r;
			f=h;
			if (i<sd-2) {
				g = surDiag[i+1];
				oldcs = cs;
				oldsn = sn;
			}
		}

		surDiag[sd-2] = h*sn;
		D[sd-1] = h*cs;
	}

	// switch U and V if needed :
	if (transposeA) {
		SWAP(U,V);
		A = A.transpose(); // useful to get back right sizes
	}

	// reorganize the values to have nice positive decreasing diagonal
	SVD_reorder(U,V,D);
	return;
}

void SVD_reorder(Matrice& U, Matrice& V, Vecteur& D) {
	// Given the output of decompose, this routine sorts the singular values, and corresponding lines/columns
	// of U and V, by decreasing magnitude. Also, signs in D are flipped so as to get positive eigenvalues.

	Int sd=D.nbElems; // sd as "size (of) diagonal"
	Int* permut=(Int*)malloc(sd*sizeof(Int)); // permutation array
	// initialized with identity :
	for (Int i=sd-1; i>=0; i--) permut[i]=i;
	// simple sort by insertion

	for (Int i=1; i<sd; i++) {
		Int j=i-1;
		while (j>=0 && fabs(D[i])>fabs(D[j])) j--;
		if (j < i-1) {
			j++;
			Real valDiag=D[i]; // store
			Int valPerm=permut[i]; // ..change permutation too !
			// shift :
			for (Int k=i; k>j; k--) {
				D[k]=D[k-1];
				permut[k]=permut[k-1];
			}
			// insert :
			D[j]=valDiag;
			permut[j]=valPerm;
		}
	}

	// rearrange columns by following permutation cycles
	bool* filled=(bool*)calloc(sd,sizeof(bool));
	Int nbRows=U.nbRows(), nbCols=V.nbCols();
	Vecteur tmpVect_U(nbRows);
	Vecteur tmpVect_V(nbCols);
	Int count=sd-1, index;
	while (count>=0) {
		while (count>=0 && filled[count]) count--;
		if (count<0) break;
		index=-1;
		while (!filled[count]) {
			if (index>=0) {
				// fill in one column and store the old one :
				for (Int i=nbRows-1; i>=0; i--)
					SWAP(U[i][index],tmpVect_U[i]);
				for (Int i=nbCols-1; i>=0; i--)
					SWAP(V[i][index],tmpVect_V[i]);
				filled[index]=true;
			}
			else { // the first visited index : we just store vectors in tmpVect
				index=count;
				tmpVect_U = U.getCol(index);
				tmpVect_V = V.getCol(index);
			}
			index=permut[index];
		}
	}
	free(filled);
	free(permut);

	// flip signs :
	for (Int i=sd-1; i>=0; i--) {
		if (D[i]<0) {
			D[i]=-D[i];
			for (Int j=nbRows-1; j>=0; j--) U[j][i]=-U[j][i];
		}
	}
}
