#ifndef SINGULARDECOMP_H_INCLUDED
#define SINGULARDECOMP_H_INCLUDED

#include "AllUsefulHeaders.hpp"
#include "Vecteur.hpp"
#include "Matrice.hpp"

#define ZERO_EPS 1.0e-7

// solves X beta ~= Y with SVD for least squares
Matrice solveLinearReg(Matrice& X,Matrice& Y);

// a function that compute a rotation to apply to cancel one dimension:
inline void rot(Real f, Real g, Real& cs, Real& sn, Real& r) {
	if (fabs(f) < ZERO_EPS) {
		cs=0.0;
		sn=1.0;
		r=g;
	}
	else if (fabs(f)>fabs(g)) {
		Real t=g/f;
		Real tt=pythag(1,t);
		cs=1/tt;
		sn = t * cs;
		r = f * tt;
	}
	else {
		Real t=f/g;
		Real tt=pythag(1,t);
		sn=1/tt;
		cs = t * sn;
		r = g * tt;
	}
}

// 4 subroutines to simplify the code:
inline void findLeftHouseholdVect(Matrice& A, Vecteur& vCol, Int indLine, Int indCol) {
	// check if column should be treated
	Int indNZ=indLine+1;
	Int nbRows = A.nbRows();
	while (indNZ < nbRows && fabs(A[indNZ][indCol]) < ZERO_EPS) indNZ++;
	if (indNZ >= nbRows)
		// code means: "identity matrix"
		vCol[0]=INFTY_REAL;

	else {
		Real alpha=0.0;
		if (indLine>0) {
			Vecteur tmpVect(A.getCol(indCol));
			tmpVect[indLine-1]=0.0;
			alpha = tmpVect.norm2();
		}
		else alpha = A.getCol(indCol).norm2();

		if (A[indLine][indCol]>0.0) alpha=-alpha;
		vCol[0]=A[indLine][indCol]+alpha;
		for (Int i=1; i<nbRows-indLine; i++) vCol[i]=A[indLine+i][indCol];
		// not normalized to gain in stability
	}
}

inline void findRightHouseholdVect(Matrice& A, Vecteur& vLine, Int indLine, Int indCol) {
	// check if line should be treated
	Int indNZ=indCol+1;
	Int nbCols = A.nbCols();
	while (indNZ < nbCols && fabs(A[indLine][indNZ]) < ZERO_EPS) indNZ++;
	if (indNZ >= nbCols)
		// code means: "identity matrix"
		vLine[0]=INFTY_REAL;

	else {
		Vecteur tmpVect(A[indLine]);
		tmpVect[indCol-1]=0.0;
		Real alpha = tmpVect.norm2();

		if (A[indLine][indCol]>0.0) alpha=-alpha;
		vLine[0]=A[indLine][indCol]+alpha;
		for (Int i=1; i<nbCols-indCol; i++) vLine[i]=A[indLine][indCol+i];
		// not normalized to gain in stability
	}
}

inline void applyLeftTransformation(Matrice& A, Matrice& U, Vecteur& vCol, Int indLine) {
	// part of HouseHolder matrix stored to fasten computations:
    if (vCol[0]==INFTY_REAL)
	    // just nothing to do : identity transformation
	    return;

	vCol.normalize2();

    // U <-- U * tr(H)=H
	U.rightMultByH(vCol);

	// ..also apply transformation to matrix A <-- H * A
	A.leftMultByH(vCol);
}

inline void applyRightTransformation(Matrice& A, Matrice& V, Vecteur& vLine, Int indCol) {
	// part of HouseHolder matrix stored to fasten computations:
    if (vLine[0]==INFTY_REAL)
	    // just nothing to do : identity transformation
	    return;

	vLine.normalize2();

	// right transformation to matrix V <-- V * tr(H)=H
	V.rightMultByH(vLine);

	// ..also apply transformation to matrix A <-- A * tr(H)=H
	A.rightMultByH(vLine);
}

// decomposition SVD :
void SVD_decomp(Matrice& A, Matrice& U, Matrice& V, Vecteur& D);

void SVD_reorder(Matrice& U, Matrice& V, Vecteur& D);

#endif // SINGULARDECOMP_INCLUDED
